//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/UIntProperty.h
//! @brief     Defines class UIntProperty
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DESCRIPTOR_UINTPROPERTY_H
#define BORNAGAIN_GUI_MODEL_DESCRIPTOR_UINTPROPERTY_H

#include "Fit/Param/RealLimits.h"
#include "GUI/Model/Descriptor/UIntDescriptor.h"
#include <variant>

class Streamer;
class SessionItem;

//! Class for representing a uint value, its attributes and its accessors.
//!
//! For a complete description please refer to the documentation of DoubleProperty. Everything said
//! there also applies to this class, of course not for double values but for uint values.
//!
//! \sa DoubleProperty
//! \sa UIntDescriptor
//!
class UIntProperty {
public:
    void init(const QString& label, const QString& tooltip, uint value,
              const std::variant<QString, Unit>& unit, const QString& persistentTag);
    void init(const QString& label, const QString& tooltip, uint value,
              const std::variant<QString, Unit>& unit, const RealLimits& limits,
              const QString& persistentTag);

    UIntDescriptor descriptor() const { return m_descriptor; }
    operator UIntDescriptor() const { return m_descriptor; }
    operator uint() const { return m_value; }
    void set(uint d) { m_value = d; }
    uint get() const { return m_value; }

    QString persistentTag() const { return m_persistentTag; }
    QString uid() const { return m_uid; }
    void setUid(const QString& uid) { m_uid = uid; }

private:
    uint m_value = 0;
    QString m_persistentTag;
    QString m_uid;

    UIntDescriptor m_descriptor;
};

namespace Serialize {
void rwProperty(Streamer& s, UIntProperty& d);
} // namespace Serialize

//! Add a member, a getter and a bunch of overloaded setters for a UIntProperty
#define UINT_PROPERTY(nameLower, nameUpper)                                                        \
private:                                                                                           \
    UIntProperty m_##nameLower;                                                                    \
                                                                                                   \
public:                                                                                            \
    UIntDescriptor nameLower() const                                                               \
    {                                                                                              \
        return m_##nameLower;                                                                      \
    }                                                                                              \
    void set##nameUpper(int v)                                                                     \
    {                                                                                              \
        m_##nameLower.set(v);                                                                      \
    }                                                                                              \
    void set##nameUpper(uint v)                                                                    \
    {                                                                                              \
        m_##nameLower.set(v);                                                                      \
    }                                                                                              \
    void set##nameUpper(size_t v)                                                                  \
    {                                                                                              \
        m_##nameLower.set((int)v);                                                                 \
    }


#endif // BORNAGAIN_GUI_MODEL_DESCRIPTOR_UINTPROPERTY_H
