//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ParticleCompositionItem.cpp
//! @brief     Implements class ParticleCompositionItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/CatSample/ItemWithParticlesCatalog.h"
#include "GUI/Model/CatSample/RotationItemCatalog.h"
#include "GUI/Model/Sample/MesocrystalItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Support/XML/Serialize.h"
#include "Sample/Particle/Mesocrystal.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Particle/ParticleCoreShell.h"
#include "Sample/Scattering/Rotations.h"

namespace {

const QString abundance_tooltip = "Proportion of this type of particles normalized to the \n"
                                  "total number of particles in the layout";

const QString position_tooltip = "Relative position of the particle's reference point \n"
                                 "in the coordinate system of the parent (nm)";

} // namespace


ParticleCompositionItem::ParticleCompositionItem(const MaterialItems* materials)
    : ItemWithParticles(abundance_tooltip, position_tooltip)
    , m_materials(materials)
{
    ASSERT(m_materials);
}

void ParticleCompositionItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, m_abundance);
    Serialize::rwProperty(s, m_position);
    Serialize::rwSelected<RotationItemCatalog>(s, m_rotation);
    Serialize::rwCatalogized<ItemWithParticlesCatalog>(s, "Particles", m_particles, m_materials);
}

std::unique_ptr<ParticleComposition> ParticleCompositionItem::createParticleComposition() const
{
    auto P_composition = std::make_unique<ParticleComposition>();
    P_composition->setAbundance(abundance());
    for (auto* p : m_particles) {
        if (auto* particleItem = dynamic_cast<ParticleItem*>(p)) {
            if (auto P_particle = particleItem->createParticle())
                P_composition->addParticle(*P_particle);
        } else if (auto* coreShellItem = dynamic_cast<ParticleCoreShellItem*>(p)) {
            if (auto P_particle_coreshell = coreShellItem->createParticleCoreShell())
                P_composition->addParticle(*P_particle_coreshell);
        } else if (auto* compositionItem = dynamic_cast<ParticleCompositionItem*>(p)) {
            if (auto P_child_composition = compositionItem->createParticleComposition())
                P_composition->addParticle(*P_child_composition);
        } else if (auto* mesocrystalItem = dynamic_cast<MesocrystalItem*>(p)) {
            if (auto P_meso = mesocrystalItem->createMesocrystal())
                P_composition->addParticle(*P_meso);
        }
    }

    P_composition->setParticlePosition(position());
    if (const auto r = createRotation(); r && !r->isIdentity())
        P_composition->setRotation(*r);

    return P_composition;
}

QVector<ItemWithParticles*> ParticleCompositionItem::particles() const
{
    return m_particles;
}

void ParticleCompositionItem::addParticle(ItemWithParticles* particle)
{
    m_particles << particle;
}

void ParticleCompositionItem::removeParticle(ItemWithParticles* particle)
{
    m_particles.removeAll(particle);
    delete particle;
}

QVector<ItemWithParticles*> ParticleCompositionItem::containedItemsWithParticles() const
{
    QVector<ItemWithParticles*> result;
    for (auto* particle : m_particles)
        result << particle << particle->containedItemsWithParticles();
    return result;
}
